//! This module contains a [`Padding`] setting of a cell on a [`Table`].
//!
//! # Example
//!
//! ```
//! use tabled::{TableIteratorExt, Padding, Style, Modify, object::Cell};
//!
//! let data = "2022".chars();
//!
//! let table = data.table()
//!     .with(Style::modern())
//!     .with(Modify::new(Cell(2, 0)).with(Padding::new(1, 1, 2, 2)))
//!     .to_string();
//!
//! assert_eq!(
//!     table,
//!     concat!(
//!         "┌──────┐\n",
//!         "│ char │\n",
//!         "├──────┤\n",
//!         "│ 2    │\n",
//!         "├──────┤\n",
//!         "│      │\n",
//!         "│      │\n",
//!         "│ 0    │\n",
//!         "│      │\n",
//!         "│      │\n",
//!         "├──────┤\n",
//!         "│ 2    │\n",
//!         "├──────┤\n",
//!         "│ 2    │\n",
//!         "└──────┘",
//!     ),
//! );
//! ```
//!
//! [`Table`]: crate::Table

use papergrid::{Entity, Grid, Indent};

use crate::CellOption;

#[cfg(feature = "color")]
use crate::style::Color;

/// Padding is responsible for a left/right/top/bottom inner indent of a particular cell.
///
/// ```rust,no_run
/// # use tabled::{Style, Padding, object::Rows, Table, Modify};
/// # let data: Vec<&'static str> = Vec::new();
/// let table = Table::new(&data).with(Modify::new(Rows::single(0)).with(Padding::new(0, 0, 1, 1).set_fill('>', '<', '^', 'V')));
/// ```
#[derive(Debug)]
pub struct Padding(papergrid::Padding);

impl Padding {
    /// Construct's an Padding object.
    ///
    /// It uses space(' ') as a default fill character.
    /// To set a custom character you can use [`Self::set_fill`] function.
    pub fn new(left: usize, right: usize, top: usize, bottom: usize) -> Self {
        Self(papergrid::Padding {
            top: Indent::spaced(top),
            bottom: Indent::spaced(bottom),
            left: Indent::spaced(left),
            right: Indent::spaced(right),
        })
    }

    /// Construct's an Padding object with all sides set to 0.
    ///
    /// It uses space(' ') as a default fill character.
    /// To set a custom character you can use [`Self::set_fill`] function.
    pub fn zero() -> Self {
        let indent = Indent::spaced(0);
        Self(papergrid::Padding {
            top: indent,
            bottom: indent,
            left: indent,
            right: indent,
        })
    }

    /// The function, sets a characters for the padding on an each side.
    pub fn set_fill(mut self, left: char, right: char, top: char, bottom: char) -> Self {
        self.0.left.fill = left;
        self.0.right.fill = right;
        self.0.top.fill = top;
        self.0.bottom.fill = bottom;
        self
    }
}

impl CellOption for Padding {
    fn change_cell(&mut self, grid: &mut Grid, entity: Entity) {
        grid.set_padding(entity, self.0);
    }
}

/// List of colors for a [Padding].
///
/// ```rust,no_run
/// # use tabled::{padding::PaddingColor, style::Color, Table, Modify, object::Segment};
/// # use owo_colors::OwoColorize;
/// # use std::convert::TryFrom;
/// # let data: Vec<&'static str> = Vec::new();
/// let table = Table::new(&data)
///     .with(Modify::new(Segment::all()).with(PaddingColor::new(
///         Color::try_from(" ".on_blue().red().to_string()).unwrap(),
///         Color::try_from(" ".red().to_string()).unwrap(),
///         Color::try_from(" ".on_blue().red().to_string()).unwrap(),
///         Color::try_from(" ".on_green().to_string()).unwrap(),
///     )));
/// ```
#[cfg(feature = "color")]
#[cfg_attr(docsrs, doc(cfg(feature = "color")))]
#[derive(Debug, Clone, Default)]
pub struct PaddingColor {
    inner: papergrid::PaddingColor,
}

#[cfg(feature = "color")]
impl PaddingColor {
    /// Creates a new [PaddingColor] with colors set for all sides.
    pub fn new(top: Color, bottom: Color, left: Color, right: Color) -> Self {
        Self {
            inner: papergrid::PaddingColor {
                bottom,
                left,
                right,
                top,
            },
        }
    }
}

#[cfg(feature = "color")]
impl CellOption for PaddingColor {
    fn change_cell(&mut self, grid: &mut Grid, entity: Entity) {
        grid.set_padding_color(entity, self.inner.clone());
    }
}
